# Copyright (C) 2005 Micha Lenk <micha@lenk.info>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# This file is based on the parent class file MemberAdaptor.py, so portions are
# Copyright (C) 2001-2003 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""PostgreSQL Mailman membership adaptor.

This adaptor gets and sets member information on the MailList object given to
the constructor. It reads and stores it's membership information in a
PostgreSQL database.

Configuration is done in the site-wide mm_pgsql_cfg.py. List-wise
configuration can be done in the pgsql.py file in the list directory.
"""

# We depend on the availability of the site packages (for postgresql), and as
# MailMan is very selective about the included module paths (and likes to run
# Python with the '-S' option) we need to load it explicitly.
import site

import os
import time
import errno
import base64
import cPickle
import weakref
from Mailman import mm_cfg
from Mailman import mm_pgsql_cfg
from Mailman import Utils
from Mailman import Site
from Mailman import MemberAdaptor
from Mailman.Logging.Syslog import syslog
from Mailman.Errors import *
from pyPgSQL import PgSQL, libpq

# Delivery statuses
ENABLED  = MemberAdaptor.ENABLED                  # enabled
UNKNOWN  = MemberAdaptor.UNKNOWN                  # legacy disabled
BYUSER   = MemberAdaptor.BYUSER                   # disabled by user choice
BYADMIN  = MemberAdaptor.BYADMIN                  # disabled by admin choice
BYBOUNCE = MemberAdaptor.BYBOUNCE                 # disabled by bounces


# Structure for PostgreSQL configuration
class pg:
    pass


class PostgreSQLMemberships(MemberAdaptor.MemberAdaptor):
    def __init__(self, mlist):
        self.internal_name = mlist.internal_name()
        self.__mlist = weakref.ref(mlist)
        
        # Read list specific configuration
	filename = os.path.join(Site.get_listpath(mlist.internal_name()), 'pgsql.py')
        syslog('debug', 'Initialise %s: Reading listspecific SQL settings from %s',
			self.internal_name, filename)
        dict = {}
        try:
            execfile(filename, dict)
        except IOError, e:
            # Ignore missing files, but log other errors
            if e.errno == errno.ENOENT:
                pass
            else:
                syslog('error', 'IOError reading list\'s SQL configuration: %s', e)
        for varname in ('DSN', 'DATABASE', 'USER', 'PASSWORD', 'TABLE', \
                        'LISTCONDITION', 'FIELD_LISTNAME', 'FIELD_MAILADDR', \
			'FIELD_KEY', 'FIELD_PASSWORD', 'FIELD_DIGEST', \
			'FIELD_LANGUAGE', 'FIELD_OPTIONS', 'FIELD_NAME', \
			'FIELD_DS', 'FIELD_DSTIME', 'FIELD_BOUNCE_INFO'):
            if (dict.has_key(varname)):
                pg.__dict__[varname] = dict.get(varname)
            else:
                pg.__dict__[varname] = mm_pgsql_cfg.__dict__.get(varname)
	# Replace placeholder <LISTNAME> by current list's name
	pg.LISTCONDITION = pg.LISTCONDITION.replace('<FIELD_LISTNAME>', pg.FIELD_LISTNAME)
	pg.LISTCONDITION = pg.LISTCONDITION.replace('<LISTNAME>', self.internal_name)

        # Initialise database connection
        try:
            self.__conn = PgSQL.connect(dsn=pg.DSN, \
                                    database=pg.DATABASE, \
                                    user=pg.USER, \
                                    password=pg.PASSWORD, \
                                    client_encoding="utf-8")
        except libpq.DatabaseError, e:
            syslog('error', 'Couldn\'t connect to database: %s', e)
            self.__conn = None
            self.__db_error = e

    #
    # The readable interface
    #
    def __get_SQL_column(self, query):
        """Get the first column of an SQL query."""
        if self.__conn == None: raise libpq.DatabaseError, self.__db_error
        cursor = self.__conn.cursor()
	cursor.execute(query)
        row = cursor.fetchone()
        column = []
        while (row <> None):
            column.append(row[0])
            row = cursor.fetchone()
        cursor.close()
        return column

    def __get_LCE_list(self, query):
        members = self.__get_SQL_column(query)
        lc_members = []
        for a_member in members:
            lc_members.append(a_member.lower())
        return lc_members

    def __getMemberValue(self, member, field):
        """Get the value out of the database of the given field for the given member.

        If member does not refer to a valid member, raise NotAMemberError.
        """
        if self.__conn == None: raise libpq.DatabaseError, self.__db_error
        query = "SELECT %s FROM %s WHERE %s AND %s='%s'" % \
                        (field, pg.TABLE, pg.LISTCONDITION, \
                        pg.FIELD_KEY, self.getMemberKey(member))
        if field.upper() not in ['NULL', 'TRUE', 'FALSE'] and field[0].isalpha():
	    query += ' ORDER BY ' + field
	cursor = self.__conn.cursor()
        cursor.execute(query)
        row = cursor.fetchone()
        cursor.close()
        if (row == None): raise NotAMemberError, member
        return row[0]

    def __setMemberValue(self, member, field, newvalue):
        """Set the value of the given field for the given member.

        If member does not refer to a valid member, raise NotAMemberError.
        """
	if not self.isMember(member): raise NotAMemberError, member
	query = "UPDATE %s SET %s='%s' WHERE %s and %s='%s'" % \
	         (pg.TABLE, field, newvalue, pg.LISTCONDITION, \
		  pg.FIELD_KEY, self.getMemberKey(member))
	cursor = self.__conn.cursor()
	cursor.execute(query)
	cursor.close()
	self.__conn.commit()
    
    def getMembers(self):
        """Get the LCE for all the members of the mailing list."""
        query = "SELECT DISTINCT %s FROM %s WHERE %s" % \
                        (pg.FIELD_MAILADDR, pg.TABLE, pg.LISTCONDITION)
        return self.__get_LCE_list(query)

    def getRegularMemberKeys(self):
        """Get the LCE for all regular delivery members (i.e. non-digest)."""
        query = "SELECT DISTINCT %s FROM %s WHERE %s AND %s=FALSE" % \
                        (pg.FIELD_MAILADDR, pg.TABLE, pg.LISTCONDITION, \
                        pg.FIELD_DIGEST)
        return self.__get_LCE_list(query)

    def getDigestMemberKeys(self):
        """Get the LCE for all digest delivery members."""
        query = "SELECT DISTINCT %s FROM %s WHERE %s AND %s=TRUE" % \
                        (pg.FIELD_MAILADDR, pg.TABLE, pg.LISTCONDITION, \
                        pg.FIELD_DIGEST)
        return self.__get_LCE_list(query)

    def isMember(self, member):
        """Return 1 if member KEY/LCE is a valid member, otherwise 0."""
        result = 1
        try:
            mailaddr = self.__getMemberValue(member, pg.FIELD_MAILADDR)
        except NotAMemberError:
            result = 0
        return result

    def getMemberKey(self, member):
        """Return the KEY for the member KEY/LCE.

        If member does not refer to a valid member, raise NotAMemberError.
        """
        cursor = self.__conn.cursor()
        query = "SELECT %s FROM %s WHERE %s AND %s='%s'" % \
                            (pg.FIELD_KEY, pg.TABLE, pg.LISTCONDITION, \
                             'lower('+pg.FIELD_MAILADDR+')', member)
        try:
            cursor.execute(query)
            key = cursor.fetchone()
	    if key is not None:
	        key = str(key.get(pg.FIELD_KEY))
        except libpq.OperationalError, e:
	    syslog('debug', 'getMemberKey(%s): A: SQL-%s' % (member, e))
	    key = None
	if key is None:
	    try:
                query = "SELECT %s FROM %s WHERE %s AND %s='%s'" % \
                        (pg.FIELD_KEY, pg.TABLE, pg.LISTCONDITION, \
                         pg.FIELD_KEY, str(member).lower())
                cursor.execute(query)
                key = cursor.fetchone()
	        if key is not None:
	            key = str(key.get(pg.FIELD_KEY))
	    except libpq.OperationalError, e:
	        syslog('debug', 'getMemberKey(%s): B: SQL-%s' % (member, e))
	        key = None
        cursor.close()
        if (key == None): raise NotAMemberError, member
        return key

    def getMemberCPAddress(self, member):
        """Return the CPE for the member KEY/LCE.

        If member does not refer to a valid member, raise NotAMemberError.
        """
        return self.__getMemberValue(member, pg.FIELD_MAILADDR)

    def getMemberCPAddresses(self, members):
        """Return a sequence of CPEs for the given sequence of members.

        The returned sequence will be the same length as members.  If any of
        the KEY/LCEs in members does not refer to a valid member, that entry
        in the returned sequence will be None (i.e. NotAMemberError is never
        raised).
        """
        result = []
        for member in members:
            try:
                cpe = self.getMemberCPAddress(member)
            except NotAMember:
                cpe = None
            result.append(cpe)
        return result

    def authenticateMember(self, member, response):
        """Authenticate the member KEY/LCE with the given response.

        If the response authenticates the member, return a secret that is
        known only to the authenticated member.  This need not be the member's
        password, but it will be used to craft a session cookie, so it should
        be persistent for the life of the session.

        If the authentication failed return False.  If member did not refer to
        a valid member, raise NotAMemberError.

        Normally, the response will be the password typed into a web form or
        given in an email command, but it needn't be.  It is up to the adaptor
        to compare the typed response to the user's authentication token.
        """
	if (pg.FIELD_PASSWORD.upper() == 'NULL'): return False
        if response == self.getMemberPassword(member): return True
        return False

    def getMemberPassword(self, member):
        """Return the member's password.

        If the member KEY/LCE is not a member of the list, raise
        NotAMemberError.
        """
        return self.__getMemberValue(member, pg.FIELD_PASSWORD)

    def getMemberLanguage(self, member):
        """Return the preferred language for the member KEY/LCE.

        The language returned must be a key in mm_cfg.LC_DESCRIPTIONS and the
        mailing list must support that language.

        If member does not refer to a valid member, the list's default
        language is returned instead of raising a NotAMemberError error.
        """
	mlist = self.__mlist()
	if mlist is None:
		mlist = MailList(self.internal_name)
		self.__mlist = weakref.ref(mlist)
        try:
            lang = self.__getMemberValue(member, pg.FIELD_LANGUAGE)
            if lang not in mlist.available_languages:
                lang = mlist.preferred_language
        except NotAMemberError:
            lang = mlist.preferred_language
        return lang

    def getMemberOption(self, member, flag):
        """Return the boolean state of the member option for member KEY/LCE.

        Option flags are defined in Defaults.py.

        If member does not refer to a valid member, raise NotAMemberError.
        """
        option = self.__getMemberValue(member, pg.FIELD_OPTIONS)
        return not not (option & flag)

    def getMemberName(self, member):
        """Return the full name of the member KEY/LCE.

        None is returned if the member has no registered full name.  The
        returned value may be a Unicode string if there are non-ASCII
        characters in the name.  NotAMemberError is raised if member does not
        refer to a valid member.
        """
        return self.__getMemberValue(member, pg.FIELD_NAME).decode('utf8')

    def getMemberTopics(self, member):
        """Return the list of topics this member is interested in.

        The return value is a list of strings which name the topics.
        """
        # TODO: really implement it XXX
        return []

    def getDeliveryStatus(self, member):
        """Return the delivery status of this member.

        Value is one of the module constants:

            ENABLED  - The deliveries to the user are not disabled
            UNKNOWN  - Deliveries are disabled for unknown reasons.  The
                       primary reason for this to happen is that we've copied
                       their delivery status from a legacy version which didn't
                       keep track of disable reasons
            BYUSER   - The user explicitly disable deliveries
            BYADMIN  - The list administrator explicitly disabled deliveries
            BYBOUNCE - The system disabled deliveries due to bouncing

        If member is not a member of the list, raise NotAMemberError.
        """
        status = self.__getMemberValue(member, pg.FIELD_DS)
        if (status <> ENABLED and status <> BYUSER and \
            status <> BYADMIN and status <> BYBOUNCE): status = UNKNOWN
        return status

    def getDeliveryStatusChangeTime(self, member):
        """Return the time of the last disabled delivery status change.

        If the current delivery status is ENABLED, the status change time will
        be zero.  If member is not a member of the list, raise
        NotAMemberError.
        """
        return self.__getMemberValue(member, pg.FIELD_DSTIME)

    def getDeliveryStatusMembers(self,
                                 status=(UNKNOWN, BYUSER, BYADMIN, BYBOUNCE)):
        """Return the list of members with a matching delivery status.

        Optional `status' if given, must be a sequence containing one or more
        of ENABLED, UNKNOWN, BYUSER, BYADMIN, or BYBOUNCE.  The members whose
        delivery status is in this sequence are returned.
        """
        status_members = []
        for a_member in self.getMembers():
            a_status = self.getDeliveryStatus(a_member)
            if a_status in status:
                status_members.append(a_member)
	syslog('debug', 'getDeliveryStatusMembers() ends up in: %s' % status_members)
        return status_members

    def getBouncingMembers(self):
        """Return the list of members who have outstanding bounce information.

        This list of members doesn't necessarily overlap with
        getDeliveryStatusMembers() since getBouncingMembers() will return
        member who have bounced but not yet reached the disable threshold.
        """
        query = "SELECT %s FROM %s WHERE %s AND %s!=''" % \
                        (pg.FIELD_KEY, pg.TABLE, pg.LISTCONDITION,
                        pg.FIELD_BOUNCE_INFO)
	syslog('debug', 'getBouncingMembers() SQL-Query: %s' % query)
        bouncing_members = self.__get_SQL_column(query)
	syslog('debug', 'getBouncingMembers() ends up in: %s' % bouncing_members)
	return bouncing_members

    def getBounceInfo(self, member):
        """Return the member's bounce information.

        A value of None means there is no bounce information registered for
        the member.

        Bounce info is opaque to the MemberAdaptor.  It is set by
        setBounceInfo() and returned by this method without modification.

        If member is not a member of the list, raise NotAMemberError.
        """
        bounce_info = self.__getMemberValue(member, pg.FIELD_BOUNCE_INFO)
        if bounce_info == None or bounce_info is '': return None
        return cPickle.loads(base64.decodestring(bounce_info))


    #
    # The writeable interface
    #
    def addNewMember(self, member, **kws):
        """Subscribes a new member to the mailing list.

        member is the case-preserved address to subscribe.  The LCE is
        calculated from this argument.  Return the new member KEY.

        This method also takes a keyword dictionary which can be used to set
        additional attributes on the member.  The actual set of supported
        keywords is adaptor specific, but should at least include:

        - digest == subscribing to digests instead of regular delivery
        - password == user's password
        - language == user's preferred language
        - realname == user's full name (should be Unicode if there are
          non-ASCII characters in the name)

        Any values not passed to **kws is set to the adaptor-specific
        defaults.

        Raise AlreadyAMemberError it the member is already subscribed to the
        list.  Raises ValueError if **kws contains an invalid option.
        """
	mlist = self.__mlist()
	if mlist is None:
		mlist = MailList(self.internal_name)
		self.__mlist = weakref.ref(mlist)
	assert mlist.Locked()
	# Make sure this address isn't already a member
	if self.isMember(member.lower()):
		raise AlreadyAMemberError
	# initialise member attributes
	new_member = {}
	new_member['FIELD_MAILADDR'] = member
	new_member['FIELD_LISTNAME'] = self.internal_name
	new_member['FIELD_DIGEST'] = 0
	new_member['FIELD_PASSWORD'] = Utils.MakeRandomPassword()
	new_member['FIELD_LANGUAGE'] = mlist.preferred_language
	new_member['FIELD_OPTIONS'] = mlist.new_member_options
	new_member['FIELD_DS'] = ENABLED
	new_member['FIELD_DSTIME'] = 0.0
	new_member['FIELD_BOUNCE_INFO'] = ''
	# Parse the keywords
        if kws.has_key('digest'):
            new_member['FIELD_DIGEST'] = kws['digest']
            del kws['digest']
        if kws.has_key('password'):
            new_member['FIELD_PASSWORD'] = kws['password']
            del kws['password']
        if kws.has_key('language'):
            new_member['FIELD_LANGUAGE'] = kws['language']
            del kws['language']
        if kws.has_key('realname'):
            new_member['FIELD_NAME'] = kws['realname']
            del kws['realname']
        # Assert that no other keywords are present
        if kws:
            raise ValueError, kws.keys()
	# sanitycheck for fieldnames
	for key in new_member.keys():
	    if pg.__dict__.get(key).lower() in ['null', 'true', 'false'] or \
	       not pg.__dict__.get(key)[0].isalpha():
	        del new_member[key]
	# build query
	fieldname = new_member.keys()[0]
	fields = pg.__dict__.get(fieldname)
	values = "'" + new_member.get(fieldname) + "'"
	for fieldname in new_member.keys()[1:]:
	    fields += ', ' + pg.__dict__.get(fieldname)
	    values += ", '" + str(new_member.get(fieldname)) + "'"
	query = "INSERT INTO %s (%s) VALUES (%s)" % \
			(pg.TABLE, fields, values)
	# submit to databse
	cursor = self.__conn.cursor()
	cursor.execute(query)
	cursor.close()
	self.__conn.commit()

    def removeMember(self, memberkey):
        """Unsubscribes the member from the mailing list.

        Raise NotAMemberError if member is not subscribed to the list.
        """
        if not self.isMember(memberkey): raise NotAMemberError, memberkey
	query = "DELETE FROM %s WHERE %s AND %s='%s'" % \
		(pg.TABLE, pg.LISTCONDITION, pg.FIELD_KEY, self.getMemberKey(memberkey))
	cursor = self.__conn.cursor()
	cursor.execute(query)
	cursor.close()
	self.__conn.commit()

    def changeMemberAddress(self, memberkey, newaddress, nodelete=0):
        """Change the address for the member KEY.

        memberkey will be a KEY, not an LCE.  newaddress should be the
        new case-preserved address for the member; the LCE will be calculated
        from newaddress.

        If memberkey does not refer to a valid member, raise NotAMemberError.
        No verification on the new address is done here (such assertions
        should be performed by the caller).

        If nodelete flag is true, then the old membership is not removed.
        """
	mlist = self.__mlist()
	if mlist is None:
		mlist = MailList(self.internal_name)
		self.__mlist = weakref.ref(mlist)
        assert mlist.Locked()
        # Make sure the old address is a member.  Assertions that the new
        # address is not already a member is done by addNewMember() below.
        if not self.isMember(member): raise NotAMemberError, member
        # Get the old values
        fullname = self.getMemberName(member)
	flags = self.__getMemberValue(member, pg.FIELD_OPTIONS)
        password = self.getMemberPassword(member) or Utils.MakeRandomPassword()
        lang = self.getMemberLanguage(member)
        # First, possibly delete the old member
        if not nodelete:
            self.removeMember(member)
        # Now, add the new member
        self.addNewMember(newaddress, realname=fullname,
                          password=password, language=lang)
        # Set the entire options bitfield (including digest option)
        self.__setMemberValue(member, pg.FIELD_OPTIONS, flags)

    def setMemberPassword(self, member, password):
        """Set the password for member LCE/KEY.

        If member does not refer to a valid member, raise NotAMemberError.
        Also raise BadPasswordError if the password is illegal (e.g. too
        short or easily guessed via a dictionary attack).
        """
        self.__setMemberValue(member, pg.FIELD_PASSWORD, password)

    def setMemberLanguage(self, member, language):
        """Set the language for the member LCE/KEY.

        If member does not refer to a valid member, raise NotAMemberError.
        Also raise BadLanguageError if the language is invalid (e.g. the list
        is not configured to support the given language).
        """
	mlist = self.__mlist()
	if mlist is None:
		mlist = MailList(self.internal_name)
		self.__mlist = weakref.ref(mlist)
        if language not in mlist.available_languages:
	    raise BadLanguageError
        self.__setMemberValue(member, pg.FIELD_LANGUAGE, language)

    def setMemberOption(self, member, flag, value):
        """Set the option for the given member to value.

        member is an LCE/KEY, flag is one of the option flags defined in
        Default.py, and value is a boolean.

        If member does not refer to a valid member, raise NotAMemberError.
        """
	# TODO: Still need to remember whether to send a last digest to
	#       this member if it is disabling the digest mode
	mlist = self.__mlist()
	if mlist is None:
		mlist = MailList(self.internal_name)
		self.__mlist = weakref.ref(mlist)
	if value:
	    if not mlist.digestable:
	        raise CantDigestError
	option = self.__getMemberValue(member, pg.FIELD_OPTIONS)
	if value:
	    option |= flag
	else:
	    option &= ~flag
	self.__setMemberValue(member, pg.FIELD_OPTIONS, option)

    def setMemberName(self, member, realname):
        """Set the member's full name.

        member is an LCE/KEY and realname is an arbitrary string.  It should
        be a Unicode string if there are non-ASCII characters in the name.
        NotAMemberError is raised if member does not refer to a valid member.
        """
        self.__setMemberValue(member, pg.FIELD_NAME, realname)

    def setMemberTopics(self, member, topics):
        """Add list of topics to member's interest.

        member is an LCE/KEY and realname is an arbitrary string.
        NotAMemberError is raised if member does not refer to a valid member.
        topics must be a sequence of strings.
        """
        raise NotImplementedError

    def setDeliveryStatus(self, member, status):
        """Set the delivery status of the member's address.

        Status must be one of the module constants:

            ENABLED  - The deliveries to the user are not disabled
            UNKNOWN  - Deliveries are disabled for unknown reasons.  The
                       primary reason for this to happen is that we've copied
                       their delivery status from a legacy version which didn't
                       keep track of disable reasons
            BYUSER   - The user explicitly disable deliveries
            BYADMIN  - The list administrator explicitly disabled deliveries
            BYBOUNCE - The system disabled deliveries due to bouncing

        This method also records the time (in seconds since epoch) at which
        the last status change was made.  If the delivery status is changed to
        ENABLED, then the change time information will be deleted.  This value
        is retrievable via getDeliveryStatusChangeTime().
        """
	mlist = self.__mlist()
	if mlist is None:
		mlist = MailList(self.internal_name)
		self.__mlist = weakref.ref(mlist)
        assert mlist.Locked()
        if status == ENABLED: new_timestamp = 0
        else: new_timestamp = time.time()
        query = "UPDATE %s SET %s='%d', %s='%f' WHERE %s AND %s='%s'" % \
                        (pg.TABLE, pg.FIELD_DS, status, \
                        pg.FIELD_DSTIME, new_timestamp, \
                        pg.LISTCONDITION, \
                        pg.FIELD_KEY, self.getMemberKey(member))
        cursor = self.__conn.cursor()
        cursor.execute(query)
        cursor.close()
	self.__conn.commit()

    def setBounceInfo(self, member, info):
        """Set the member's bounce information.

        When info is None, any bounce info for the member is cleared.

        Bounce info is opaque to the MemberAdaptor.  It is set by this method
        and returned by getBounceInfo() without modification.
        """
	mlist = self.__mlist()
	if mlist is None:
		mlist = MailList(self.internal_name)
		self.__mlist = weakref.ref(mlist)
        assert mlist.Locked()
        if not self.isMember(member): raise NotAMemberError, member
        if info == None:
            q_bounce_info = 'NULL'
        else:
	    q_bounce_info = "'" + base64.encodestring(cPickle.dumps(info, cPickle.HIGHEST_PROTOCOL)) + "'"
        query = "UPDATE %s SET %s=%s WHERE %s AND %s='%s'" % \
                        (pg.TABLE, pg.FIELD_BOUNCE_INFO, q_bounce_info, \
                        pg.LISTCONDITION, \
                        pg.FIELD_KEY, self.getMemberKey(member))
        cursor = self.__conn.cursor()
        cursor.execute(query)
        cursor.close()
	self.__conn.commit()
    
    def __del__(self):
	syslog('debug', 'Uninitialising list %s', self.internal_name)
        self.__conn.commit()
	self.__conn.close()
